import markdown
from markdown.extensions.tables import TableExtension
from markdown_blockdiag import BlockdiagExtension
from rest_framework.response import Response

from bridger.settings import bridger_settings


def get_markdown_docs(docs):
    try:
        with open(docs, "r") as f:
            return Response(markdown.markdown(f.read(), extensions=bridger_settings.DEFAULT_MARKDOWN_EXTENSIONS,))
    except FileNotFoundError:
        return Response(markdown.markdown(docs, extensions=bridger_settings.DEFAULT_MARKDOWN_EXTENSIONS,))
    except AttributeError:
        return Response("No documentation available.")



from .formatting import Condition, Formatting, FormattingRule
from .instance_display import FieldSet, InstanceDisplay, Section, SectionList
from .list_display import Calendar, Field, Legend, LegendItem, ListDisplay


from rest_framework.test import APIRequestFactory
from rest_framework.request import Request
from django.contrib.auth import get_user_model
from rest_framework import status

from .utils import get_model_factory, format_number, get_data_factory_mvs, get_kwargs
from termcolor import colored
import json

class TestModelClass:
    def __init__(self, model):
        self.model = model
        self.factory = get_model_factory(model)
    
    def test_get_endpoint_basename(self):
        assert self.model.get_endpoint_basename()
        print("- TestModelClass:test_get_endpoint_basename", colored("PASSED", 'green'))

    def test_representation_endpoint(self):
        assert self.model.get_representation_endpoint()
        assert self.model.get_representation_endpoint() == self.model._meta.app_label+":"+self.model.__name__.lower()+"representation-list"
        print("- TestModelClass:test_representation_endpoint", colored("PASSED", 'green'))

    def test_representation_value_key(self):
        assert self.model.get_representation_value_key()
        assert self.model.get_representation_value_key() == "id"
        print("- TestModelClass:test_representation_value_key", colored("PASSED", 'green'))

    def test_representation_label_key(self):
        assert self.model.get_representation_label_key()
        print("- TestModelClass:test_representation_label_key", colored("PASSED", 'green'))

    def test_count_model(self):
        if self.factory is None:
            print("- TestModelClass:test_count_model", colored("WARNING - factory not found for "+self.model.__name__, 'yellow'))
        else:
            models = self.model.objects.all()
            assert models.count() == 0
            for _ in range(4):
                self.factory()
            assert models.count() == 4
            print("- TestModelClass:test_count_model", colored("PASSED", 'green')) 

    def test_str(self):
        if self.factory is None:
            print("- TestModelClass:test_str", colored("WARNING - factory not found for "+self.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            assert str(obj)
            assert isinstance(str(obj), str)
            print("- TestModelClass:test_str", colored("PASSED", 'green'))

    def test_field(self):
        if self.factory is None:
            print("- TestModelClass:test_field", colored("WARNING - factory not found for "+self.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            for field in obj._meta.get_fields():
                assert field
            print("- TestModelClass:test_field", colored("PASSED", 'green'))

    def execute_test(self):
        self.test_get_endpoint_basename()
        self.test_representation_endpoint()
        self.test_representation_value_key()
        self.test_representation_label_key()
        self.test_count_model()
        self.test_str()
        self.test_field()
    

class TestSerializerClass:
    def __init__(self, serializer):
        self.serializer = serializer
        self.factory = get_model_factory(serializer.Meta.model)
    
    def test_serializer(self):
        if self.factory is None:
            print("\n- TestSerializerClass:test_serializer:"+self.serializer.__name__, colored("WARNING - factory not found for "+self.serializer.Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            serializer = self.serializer(self.factory(), context= {'request': Request(request)})
            assert serializer.data
            print("\n- TestSerializerClass:test_serializer:"+self.serializer.__name__, colored("PASSED", 'green'))

    def execute_test(self):
        self.test_serializer()


class SuperUser:
    @classmethod
    def get_user(cls):
        try:
            superuser = get_user_model().objects.get(username="test_user")
        except (get_user_model().DoesNotExist, get_user_model().MultipleObjectsReturned):
            superuser = get_user_model().objects.create(username="test_user", password="ABC", is_active=True, is_superuser=True)
        return superuser


class TestrepresentationViewSetClass:
    def __init__(self, rmvs):
        self.rmvs = rmvs
        self.factory = get_model_factory(rmvs().get_serializer_class().Meta.model)

    # ----- LIST ROUTE TEST ----- #
    #Test representationviewset "get": "list"
    def test_representationviewset(self):
        request = APIRequestFactory().get("")
        request.user = SuperUser.get_user()
        vs = self.rmvs.as_view({"get": "list"})
        response = vs(request)
        assert response.status_code == status.HTTP_200_OK
        assert response.data
        print("\n- TestViewSetClass:test_representationviewset", colored("PASSED", 'green'))  

    
    # ----- DETAIL ROUTE TEST ----- #
    #Test representationviewset "get": "retrieve"
    def test_instancerepresentationviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_instancerepresentationviewset", colored("WARNING - factory not found for "+self.rmvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            vs = self.rmvs.as_view({"get": "retrieve"})
            obj = self.factory()
            response = vs(request, pk=obj.pk)
            assert response.status_code == status.HTTP_200_OK
            assert response.data.get('instance')
            print("- TestViewSetClass:test_instancerepresentationviewset", colored("PASSED", 'green')) 

    
    def execute_test(self, admin_client):
        # ----- LIST ROUTE TEST ----- #
        self.test_representationviewset()
        self.test_instancerepresentationviewset()
    

class TestViewSetClass:
    def __init__(self, mvs):
        self.mvs = mvs
        self.factory = get_model_factory(mvs().get_serializer_class().Meta.model)

    # test viewset Option request
    def test_option_request(self):
        if self.factory is None:
            print("- TestViewSetClass:test_option_request", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().options("")
            request.user = SuperUser.get_user()      
            kwargs = get_kwargs(self.factory(), self.mvs, request)
            vs = self.mvs.as_view({"options": "options"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK
            assert response.data.get("fields")
            assert response.data.get("identifier")
            assert response.data.get("pagination")
            # assert response.data.get("pk")
             # assert response.data.get("type")
            # assert response.data.get("filter_fields")
            # assert response.data.get("search_fields")
            # assert response.data.get("ordering_fields")
            assert response.data.get("buttons")
            assert response.data.get("display")
            assert response.data.get("titles")
            assert response.data.get("endpoints")
            # assert response.data.get("preview")
            print("\n- TestViewSetClass:test_option_request", colored("PASSED", 'green'))  


    # ----- LIST ROUTE TEST ----- #
    # Test viewset "get": "list"
    def test_viewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_viewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()      
            kwargs = get_kwargs(self.factory(), self.mvs, request)
            vs = self.mvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK
            assert response.data
            assert response.data.get("results")
            print("- TestViewSetClass:test_viewset", colored("PASSED", 'green'))  

    # Test viewset "get": "list" -> aggregation
    def test_aggregation(self, namefield=None):
        if self.factory is None:
            print("- TestViewSetClass:test_aggregation", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(self.factory(), self.mvs, request)
            vs = self.mvs.as_view({"get": "list"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_200_OK
            assert response.data.get("results")
            if not response.data.get("aggregates"):
                print("- TestViewSetClass:test_aggregation:"+self.mvs.__name__, colored("WARNING - aggregates not found in "+self.mvs.__name__, 'yellow'))
            else:
                if namefield:
                    assert response.data.get("aggregates").get(namefield)
                    assert response.data.get("aggregates").get(namefield).get("#") == format_number(self.mvs().get_serializer_class().Meta.model.objects.count())
                print("- TestViewSetClass:test_aggregation", colored("PASSED", 'green')) 

    # Test viewset "get": "list" with client and endpoint
    def test_get_client_endpointviewset(self, admin_client):
        if self.factory is None:
            print("- TestViewSetClass:test_get_client_endpointviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            obj = self.factory()
            self.mvs.kwargs = get_kwargs(obj, self.mvs, request)
            ep = self.mvs.endpoint_config_class(self.mvs, request, instance=False)
            response = admin_client.get(ep.get_endpoint())
            assert response.status_code == status.HTTP_200_OK
            assert response.data.get("results")
            print("- TestViewSetClass:test_get_client_endpointviewset", colored("PASSED", 'green'))  

    # Test viewset "post": "create"
    def test_postviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_postviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            data = get_data_factory_mvs(obj, self.mvs, delete=True)
            request = APIRequestFactory().post("", data)
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs = self.mvs.as_view({"post": "create"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_201_CREATED
            assert response.data.get('instance')
            print("- TestViewSetClass:test_postviewset", colored("PASSED", 'green')) 

    # Test viewset "post": "create" with client and endpoint
    def test_post_client_endpointviewset(self, admin_client):
        if self.factory is None:
            print("- TestViewSetClass:test_post_client_endpointviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            data = get_data_factory_mvs(obj, self.mvs, delete=True)
            request = APIRequestFactory().post("", data)
            request.user = SuperUser.get_user()
            ep = self.mvs.endpoint_config_class(self.mvs, request, instance=False)
            response = admin_client.post(ep._get_create_endpoint(),
                                json.dumps(data),
                                content_type='application/json')
            assert response.status_code == status.HTTP_201_CREATED
            assert response.data.get('instance')
            print("- TestViewSetClass:test_post_client_endpointviewset", colored("PASSED", 'green'))  

    # TODO Test viewset "delete": "destroy_multiple"
    def test_destroy_multipleviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_destroy_multipleviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().delete("")
            request.user = SuperUser.get_user()
            for _ in range(4):
                obj = self.factory()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs = self.mvs.as_view({"delete": "destroy_multiple"})
            response = vs(request, **kwargs)
            assert response.status_code == status.HTTP_204_NO_CONTENT
            # assert not self.mvs().get_serializer_class().Meta.model.objects.filter(id=obj.id).exists()
        print("- TestViewSetClass:test_destroy_multipleviewset", colored("PASSED", 'green'))  
        
            
    # TODO Test viewset "delete": "destroy_multiple" with client and endpoint
    def test_destroy_multiple_client_endpointviewset(self, admin_client):
        if self.factory is None:
            print("- TestViewSetClass:test_destroy_multiple_client_endpointviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().delete("")
            request.user = SuperUser.get_user()
            list_id = []
            for _ in range(4):
                obj = self.factory()
                list_id.append(obj.id)
            request.data = list_id
            ep = self.mvs.endpoint_config_class(self.mvs, request, instance=False)
            response = admin_client.delete(ep._get_delete_endpoint())
            # queryset = self.mvs().get_serializer_class().Meta.model.objects.filter(id__in=data)
            # destroyed = queryset.delete()
            # print(destroyed)
            assert response.status_code == status.HTTP_204_NO_CONTENT


    # Test viewset "get_list_title"
    def test_get_list_title(self):
        if self.factory is None:
            print("- TestViewSetClass:test_get_list_title", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs =  self.mvs(kwargs=kwargs)
            assert vs.title_config_class(vs, request, instance=False).get_list_title()
            print("- TestViewSetClass:test_get_list_title", colored("PASSED", 'green'))  

    # Test viewset "get_list_title"
    def test_get_instance_title(self):
        if self.factory is None:
            print("- TestViewSetClass:test_get_instance_title", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs =  self.mvs(kwargs=kwargs)
            assert vs.title_config_class(vs, request, instance=False).get_instance_title()
            print("- TestViewSetClass:test_get_instance_title", colored("PASSED", 'green'))  

    # Test viewset "get_create_title"
    def test_get_create_title(self):
        if self.factory is None:
            print("- TestViewSetClass:test_get_create_title", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs =  self.mvs(kwargs=kwargs)
            assert vs.title_config_class(vs, request, instance=False).get_create_title()
            print("- TestViewSetClass:test_get_create_title", colored("PASSED", 'green'))  

    #-------------- DETAIL ROUTE TEST ------------------#
    # Test viewset "get": "retrieve"
    def test_instanceviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_instanceviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().get("")
            request.user = SuperUser.get_user()
            obj = self.factory()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs = self.mvs.as_view({"get": "retrieve"})
            response = vs(request, **kwargs, pk=obj.pk)
            assert response.status_code == status.HTTP_200_OK
            assert response.data.get('instance')
            print("- TestViewSetClass:test_instanceviewset", colored("PASSED", 'green'))  
    
    # Test "delete": "destroy"
    def test_deleteviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_deleteviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            request = APIRequestFactory().delete("")
            request.user = SuperUser.get_user()
            obj = self.factory()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs = self.mvs.as_view({"delete": "destroy"})
            response = vs(request, **kwargs, pk=obj.pk)
            assert response.status_code == status.HTTP_204_NO_CONTENT
            assert not self.mvs().get_serializer_class().Meta.model.objects.filter(id=obj.id).exists()
            print("- TestViewSetClass:test_deleteviewset", colored("PASSED", 'green'))  

    # Test "put": "update"
    def test_uptdateviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_uptdateviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            data = get_data_factory_mvs(obj, self.mvs)
            request = APIRequestFactory().put("", data)
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs = self.mvs.as_view({"put": "update"})
            response = vs(request, **kwargs, pk=obj.pk)
            assert response.status_code == status.HTTP_200_OK
            assert response.data.get('instance')
            print("- TestViewSetClass:test_uptdateviewset", colored("PASSED", 'green'))  


    #Test "patch": "partial_update",
    def test_patchviewset(self):
        if self.factory is None:
            print("- TestViewSetClass:test_patchviewset", colored("WARNING - factory not found for "+self.mvs().get_serializer_class().Meta.model.__name__, 'yellow'))
        else:
            obj = self.factory()
            data = get_data_factory_mvs(obj, self.mvs) 
            request = APIRequestFactory().patch("", data)
            request.user = SuperUser.get_user()
            kwargs = get_kwargs(obj, self.mvs, request)
            vs = self.mvs.as_view({"patch": "partial_update"})
            response = vs(request, **kwargs, pk=obj.pk, data=data)
            assert response.status_code == status.HTTP_200_OK
            assert response.data.get('instance')
            print("- TestViewSetClass:test_patchviewset", colored("PASSED", 'green'))  

    # TODO Test "get": "history_list" ??
    # TODO Test "get": "history_retrieve" ??
    # TODO 'get': 'highlight' ??

    def execute_test(self, admin_client, aggregates=None):
        self.test_option_request()
        # ----- LIST ROUTE TEST ----- #
        self.test_viewset()
        self.test_aggregation(aggregates)
        self.test_get_client_endpointviewset(admin_client)
        self.test_postviewset()
        self.test_post_client_endpointviewset(admin_client)
        self.test_destroy_multipleviewset()
        self.test_destroy_multiple_client_endpointviewset(admin_client)
        self.test_get_list_title()
        self.test_get_instance_title()
        self.test_get_create_title()
        # # ----- DETAIL ROUTE TEST ----- #
        self.test_instanceviewset()
        self.test_deleteviewset()
        self.test_uptdateviewset()
        self.test_patchviewset()
        #Test "get": "history_list"
        #Test "get": "history_retrieve"


from prettytable import PrettyTable
from decimal import Decimal


def prettytable_from_querydict(values, float_format=False):
    ptable = PrettyTable()
    keys = [key for key in values[0].keys()]
    
    ptable.field_names = list(keys)

    if float_format:
        for _values in values:
            formatted_values = list()
            for v in _values.values():
                if type(v) in [float, int, Decimal]:
                    v = f"{v:,.2f}"
                formatted_values.append(v)
            ptable.add_row(formatted_values)
    else:
        for _values in values:
            ptable.add_row(_values.values())

    print(ptable)


def prettytable_from_list(headers, values):
    ptable = PrettyTable()
    
    ptable.field_names = list(headers)

    for _values in values:
        ptable.add_row(_values.values())

    print(ptable)


from django.apps import apps
import importlib 
import factory
from rest_framework.test import APIRequestFactory
from rest_framework.request import Request
from django.urls import reverse
import json

from functools import partial
from typing import Any, Dict

from factory import Factory
from factory.base import StubObject

def get_all_subclasses(klass):
    for subclass in klass.__subclasses__():
        yield subclass
        yield from get_all_subclasses(subclass)
        
def get_model_factory(model):
    mf = [cls for cls in factory.django.DjangoModelFactory.__subclasses__() if cls._meta.model == model]
    if mf:
        return mf[0]
    return None

def get_data_factory_mvs(obj, mvs, delete=False):
    request = APIRequestFactory().get("")
    serializer = mvs().serializer_class(obj, context= {'request': Request(request)})

    fields_models = [m.name for m in mvs().serializer_class.Meta.model._meta.get_fields()]
    data = {}
    for key, value in serializer.data.items():
        if key in fields_models:
            data[key] = value
    # delete object for create with post method
    if delete:
        mvs().serializer_class.Meta.model.objects.filter(pk = obj.pk).delete()
    return data

def get_kwargs(obj, mvs, request):
    fields_models = [m for m in mvs().serializer_class.Meta.model._meta.get_fields()]         
    serializer = mvs().serializer_class(obj, context= {'request': request})
    kwargs={}
    for field in fields_models:
        if field.is_relation and field.name in serializer.data.keys():
            kwargs[field.name+"_id"] = serializer.data[field.name]
    return kwargs


def generate_dict_factory(factory: Factory):
    def convert_dict_from_stub(stub: StubObject) -> Dict[str, Any]:
        stub_dict = stub.__dict__
        for key, value in stub_dict.items():
            if isinstance(value, StubObject):
                stub_dict[key] = convert_dict_from_stub(value)
        return stub_dict

    def dict_factory(factory, **kwargs):
        stub = factory.stub(**kwargs)
        stub_dict = convert_dict_from_stub(stub)
        return stub_dict

    return partial(dict_factory, factory)

def format_number(number, is_pourcent=False, decimal=2):
    number = number if number else 0
    return f'{number:,.{decimal}{"%" if is_pourcent else "f"}}'


from datetime import date, datetime, timedelta
from django.conf import settings
from django.utils.timezone import localdate
from dateutil import rrule
from django.utils.dateparse import parse_date
from collections import defaultdict

def get_date_interval_from_get(request):
    """
    Parses a request and returns the start and end date from it.

    Parameters
    ----------
    request: The GET Request Object

    Returns
    -------
    Return a tuple in the form of (start_date, end_date). If either the start date or the end date is not present in the request None is returned in the tuple
    """
    print(request)
    start_identifier = getattr(settings, "BRIDGER_START_IDENTIFIERS", ["start", "start_date", "from", "date_gte"],)
    end_identifier = getattr(settings, "BRIDGER_END_IDENTIFIERS", ["end", "end_date", "to", "date_lte"])
    date_format = getattr(settings, "BRIDGER_DATE_FORMAT", "%Y-%m-%d")

    assert isinstance(start_identifier, list)
    assert isinstance(end_identifier, list)
    assert isinstance(date_format, str)

    start = next((identifier for identifier in start_identifier if identifier in request.GET), None,)
    end = next((identifier for identifier in end_identifier if identifier in request.GET), None)

    if start:
        try:
            start = datetime.strptime(request.GET.get(start), date_format).date()
        except ValueError:
            start = None
    if end:
        try:
            end = datetime.strptime(request.GET.get(end), date_format).date()
        except ValueError:
            end = None

    return start, end


def get_quarter_from_date(d):
    return ((d.month - 1) // 3) + 1


def get_start_date_from_date(d):
    quarter = get_quarter_from_date(d)
    return date(d.year, quarter * 3 - 2, 1)


def get_end_date_from_date(d):
    quarter = get_quarter_from_date(d)
    return date(d.year + ((quarter * 3 + 1) // 12), (quarter * 3 + 1) % 12, 1) - timedelta(days=1)


def get_start_and_end_date_from_date(d):
    return get_start_date_from_date(d), get_end_date_from_date(d)


def current_quarter_date_start(field=None, request=None, view=None):
     return get_start_date_from_date(localdate())

def current_quarter_date_end(field=None, request=None, view=None):
     return get_end_date_from_date(localdate())

def current_quarter_date_interval(field, request, view):
    return (
        current_quarter_date_start(field, request, view),
        current_quarter_date_end(field, request, view),
    )

def current_year_date_start(field, request, view):
    d = localdate()
    return date(d.year,1,1)

def current_year_date_end(field, request, view):
    d = localdate()
    return date(d.year + 1,1,1) - timedelta(days=1)

def current_year_date_interval(field, request, view):
    return (
        current_year_date_start(field, request, view),
        current_year_date_end(field, request, view),
    )

def current_month_date_start(field, request, view):
    d = localdate()
    return date(d.year,d.month,1)

def current_month_date_end(field=None, request=None, view=None):
    d = localdate()
    if d.month == 12:
        return date(d.year, 12, 31)
    return date(d.year, d.month + 1, 1) - timedelta(days=1)

def current_month_date_interval(field, request, view):
    return (
        current_month_date_start(field, request, view),
        current_month_date_end(field, request, view),
    )

    
def get_date_interval_from_request(request, request_type="GET"):
    """
    Parses a request and returns the start and end date from it.

    Parameters
    ----------
    request: The GET Request Object

    Returns
    -------
    Return a tuple in the form of (start_date, end_date). If either the start date or the end date is not present in the request None is returned in the tuple
    """

    start_identifier = ["start", "start_date", "from", "date_gte"]
    end_identifier = ["end", "end_date", "to", "date_lte"]
    params = request.GET if request_type == "GET" else request.POST
    start = None
    end = None
    if "date" in params:
        if len(params.get("date").split(",")) == 2:
            start, end = params.get("date").split(",")
    else:
        start = next(
            (params.get(identifier) for identifier in start_identifier if identifier in params),
            None,
        )
        end = next(
            (params.get(identifier) for identifier in end_identifier if identifier in params), None
        )

    if start:
        start = parse_date(start)
    if end:
        end = parse_date(end)

    return start, end




def get_number_of_hours_between_dates(
    d1, d2, skip_weekends=True, list_public_holidays=False, hours_range=range(0,23), granularity=12
):
    def convert_days_from_hours(hours, granularity, hours_per_day):
        return int(hours/granularity)*granularity/hours_per_day
    rules = rrule.rruleset()
    
    byweekday_list = [rrule.MO, rrule.TU, rrule.WE, rrule.TH, rrule.FR]
    if not skip_weekends:
        byweekday_list.extend([rrule.SA, rrule.SU])
    
    rules.rrule(
        rrule.rrule(
            freq=rrule.HOURLY,
            byweekday=byweekday_list,
            byhour=hours_range,
            dtstart=d1,
            until=d2,
        )
    )
    if list_public_holidays:
        for holiday in list_public_holidays:
            s1 = datetime(holiday.year, holiday.month, holiday.day, 0, 0, 0)
            s2 = datetime(holiday.year, holiday.month, holiday.day, 23, 59, 59)
            rules.exrule(
                rrule.rrule(
                    rrule.HOURLY,
                    dtstart=s1,
                    until=s2
                )
            )
    dates = defaultdict(int)
    for r in list(rules):
        dates[r.date()] += 1
    return {k: convert_days_from_hours(v, granularity, len(hours_range)) for k, v in dates.items()}





from bridger.serializers.fields.choice import ChoiceField
from bridger.serializers.fields.text import CharField


# TODO: Color for each state
class FSMStatusField(CharField):

    field_type = "status"

    def __init__(self, *args, **kwargs):
        self.choices = kwargs.pop("choices")
        super().__init__(read_only=True, *args, **kwargs)

    def get_representation(self, request, field_name):
        representation = super().get_representation(request, field_name)
        representation["choices"] = list()

        for choice in self.choices:
            representation["choices"].append({"value": choice[0], "label": choice[1]})

        return representation



from enum import Enum
from typing import Union


class RequestType(Enum):
    GET = "get"
    POST = "post"
    DELETE = "delete"
    PATCH = "patch"
    PUT = "put"
    OPTIONS = "options"
    HEAD = "head"


class Button(Enum):

    # Buttons
    REFRESH = "refresh"
    NEW = "new"
    DELETE = "delete"

    # Buttons and Create Buttons
    SAVE = "save"
    SAVE_AND_CLOSE = "saveandclose"
    SAVE_AND_NEW = "saveandnew"

    # Create Buttons
    RESET = "reset"

    # Custom Buttons
    DROPDOWN = "dropdown"
    HYPERLINK = "hyperlink"
    WIDGET = "widget"
    ACTION = "action"

    @classmethod
    def buttons(cls):
        return [
            cls.REFRESH.value,
            cls.NEW.value,
            cls.DELETE.value,
            cls.SAVE.value,
            cls.SAVE_AND_CLOSE.value,
            cls.SAVE_AND_NEW.value,
        ]

    @classmethod
    def create_buttons(cls):
        return [
            cls.SAVE.value,
            cls.SAVE_AND_CLOSE.value,
            cls.SAVE_AND_NEW.value,
            cls.RESET.value,
        ]

    @classmethod
    def custom_buttons(cls):
        return [
            cls.DROPDOWN.value,
            cls.HYPERLINK.value,
            cls.WIDGET.value,
            cls.ACTION.value,
        ]


class WidgetType(Enum):
    LIST = "list"
    INSTANCE = "instance"
    CHART = "chart"


class Operator(Enum):
    EQUAL = "=="
    UNEQUAL = "!="
    GREATER = ">"
    GREATER_EQUAL = ">="
    LESS = "<"
    LESS_EQUAL = "<="
    EXISTS = "∃"


class Unit(Enum):
    FRACTION = "fr"
    REM = "rem"
    PIXEL = "px"

    def __call__(self, _value):
        return (float(_value), self.value)

    def unit(self, _value: Union[float, str, int]):
        assert isinstance(_value, (float, str, int)), f"_value needs to be one of str, float or int"

        return f"{float(_value)}{self.value}"


class AuthType(Enum):
    NONE = "NONE"
    JWT = "JWT"


class WBIcon(Enum):
    ACTION = "wb-icon-action"
    ARROWS_SWAP_1 = "wb-icon-arrows-swap-1"
    ARROWS_SWAP_2 = "wb-icon-arrows-swap-2"
    ARROW_DOWN = "wb-icon-arrow-down"
    ARROW_LEFT = "wb-icon-arrow-left"
    ARROW_RIGHT = "wb-icon-arrow-right"
    ARROW_UP = "wb-icon-arrow-up"
    BANK = "wb-icon-bank"
    BELL_EMPTY = "wb-icon-bell-inverted"
    BELL_FULL = "wb-icon-bell"
    BOLD = "wb-icon-bold"
    BOOKMARK = "wb-icon-bookmark"
    BOOKMARK_INVERTED = "wb-icon-bookmark-inverted"
    BULLETED_LIST = "wb-icon-bulleted-list"
    CAKE_1 = "wb-icon-cake-1"
    CAKE_2 = "wb-icon-cake-2"
    CAKE_3 = "wb-icon-cake-3"
    CALENDAR_1 = "wb-icon-calendar-1"
    CALENDAR_2 = "wb-icon-calendar-2"
    CALENDAR_3 = "wb-icon-calendar-3"
    CAMERA_1 = "wb-icon-camera-1"
    CAMERA_2 = "wb-icon-camera-2"
    CHART_BARS_1 = "wb-icon-chart-bars-1"
    CHART_BARS_2 = "wb-icon-chart-bars-2"
    CHART_BARS_3 = "wb-icon-chart-bars-3"
    CHART_BARS_4 = "wb-icon-chart-bars-4"
    CHART_LINE_2 = "wb-icon-chart-line-2"
    CHART_PIE = "wb-icon-chart-pie"
    CHART_PIRAMID = "wb-icon-chart-piramid"
    CHART_SWITCHES = "wb-icon-chart-switches"
    CIRCLE_CHECK = "wb-icon-circle-check"
    CIRCLE_EXCLAMATION = "wb-icon-circle-exclamation"
    CIRCLE_EXCLAMATION_INVERTED = "wb-icon-circle-exclamation-inverted"
    CIRCLE_NO = "wb-icon-circle-x-inverted"
    CIRCLE_QUESTION = "wb-icon-circle-question"
    CIRCLE_QUESTION_INVERTED = "wb-icon-circle-question-inverted"
    CIRCLE_X = "wb-icon-circle-x"
    CIRCLE_YES = "wb-icon-circle-check-inverted"
    CLIPBOARD_1 = "wb-icon-clipboard-1"
    CLIPBOARD_2 = "wb-icon-clipboard-2"
    CLIPBOARD_3 = "wb-icon-clipboard-3"
    CLIPBOARD_4 = "wb-icon-clipboard-4"
    CLIPBOARD_CHECKS_1 = "wb-icon-clipboard-checks-1"
    CLIPBOARD_CHECKS_2 = "wb-icon-clipboard-checks-2"
    CLOCK = "wb-icon-clock-2"
    CLOCK_1 = "wb-icon-clock-1"
    CLOCK_3 = "wb-icon-clock-3"
    CLOCK_ALARM = "wb-icon-clock-alarm"
    COG = "wb-icon-gear-6"
    CONTACT_BOOK = "wb-icon-contact-book"
    COPY_1 = "wb-icon-copy-1"
    COPY_2 = "wb-icon-copy-2"
    COPY_3 = "wb-icon-copy-3"
    COPY_4 = "wb-icon-copy-4"
    COPY_5 = "wb-icon-copy-5"
    COPY_6 = "wb-icon-copy-6"
    COPY_7 = "wb-icon-copy-7"
    COPY_FAT_1 = "wb-icon-copy-fat-1"
    COPY_FAT_2 = "wb-icon-copy-fat-2"
    COPY_FAT_3 = "wb-icon-copy-fat-3"
    COPY_FAT_4 = "wb-icon-copy-fat-4"
    COPY_FAT_5 = "wb-icon-copy-fat-5"
    CROSSHAIR = "wb-icon-crosshair"
    DASHBOARD = "wb-icon-dashboard-2"
    DASHBOARD_1 = "wb-icon-dashboard-1"
    DATA = "wb-icon-data-grid-3"
    DATA_GRID_1 = "wb-icon-data-grid-1"
    DATA_GRID_2 = "wb-icon-data-grid-2"
    DATA_GRID_4 = "wb-icon-data-grid-4"
    DATA_GRID_5 = "wb-icon-data-grid-5"
    DATA_LIST_2 = "wb-icon-data-list-2"
    DATA_LIST_3 = "wb-icon-data-list-3"
    DATA_LIST_4 = "wb-icon-data-list-4"
    DATA_LIST_5 = "wb-icon-data-list-5"
    DATA_LIST_6 = "wb-icon-data-list-6"
    DATA_LIST_7 = "wb-icon-data-list-7"
    DEDENT = "wb-icon-dedent"
    DISABLED = "wb-icon-disabled"
    DOCUMENT_1 = "wb-icon-document-1"
    DOCUMENT_2 = "wb-icon-document-2"
    DOCUMENT_BOOK = "wb-icon-document-book"
    DOCUMENT_GEAR = "wb-icon-document-gear"
    DOCUMENT_LOCK = "wb-icon-document-lock"
    DOLLAR = "wb-icon-dollar"
    EDIT_2 = "wb-icon-edit-2"
    EDIT_3 = "wb-icon-edit-3"
    ENVELOP = "wb-icon-envelope"
    ENVELOPE_OPEN_1 = "wb-icon-envelope-open-1"
    ENVELOPE_OPEN_2 = "wb-icon-envelope-open-2"
    ENVELOPE_OPEN_DOLLAR = "wb-icon-envelope-open-dollar"
    ENVELOP_EURO = "wb-icon-envelope-open-euro"
    ENVELOP_LOCK = "wb-icon-envelope-open-suitcase"
    ENVELOP_OPEN = "wb-icon-envelope-open-3"
    ENVELOP_PERSON = "wb-icon-envelope-open-person"
    EURO = "wb-icon-euro"
    EYE = "wb-icon-eye"
    EYE_CROSSED = "wb-icon-eye-crossed"
    FEMALE = "wb-icon-female"
    FILTERS_1 = "wb-icon-filters-1"
    FILTERS_2 = "wb-icon-filters-2"
    FILTERS_OFF_1 = "wb-icon-filters-off-1"
    FILTERS_OFF_2 = "wb-icon-filters-off-2"
    FILTERS_OFF_3 = "wb-icon-filters-off-3"
    FLOPPY_DISK = "wb-icon-floppy-disk"
    FLOPPY_DISK_CHECK = "wb-icon-floppy-disk-check"
    FLOPPY_DISK_X = "wb-icon-floppy-disk-x"
    FOLDER = "wb-icon-folder"
    FOLDERS = "wb-icon-folders"
    FOLDER_EURO = "wb-icon-folder-euro"
    FOLDER_INVERTED = "wb-icon-folder-inverted"
    FOLDER_PLUS = "wb-icon-folder-plus"
    FOLDER_PLUS_INVERTED = "wb-icon-folder-plus-inverted"
    FONT_COLOR = "wb-icon-font-color"
    GEAR_1 = "wb-icon-gear-1"
    GEAR_2 = "wb-icon-gear-2"
    GEAR_3 = "wb-icon-gear-3"
    GEAR_4 = "wb-icon-gear-4"
    GEAR_5 = "wb-icon-gear-5"
    GEAR_7 = "wb-icon-gear-7"
    GLOBE = "wb-icon-globe-1"
    GLOBE_2 = "wb-icon-globe-2"
    HANDSHAKE = "wb-icon-handshake"
    HANDS_EURO = "wb-icon-handshake-euro"
    HOME = "wb-icon-home"
    HOURGLASS_EMPTY = "wb-icon-hourglass-empty"
    HOURGLASS_FULL = "wb-icon-hourglass-full"
    HOURGLASS_MIDDLE = "wb-icon-hourglass-middle"
    IMAGE = "wb-icon-image"
    INDENT = "wb-icon-indent"
    INFO = "wb-icon-circle-info"
    ITALIC = "wb-icon-italic"
    LINK = "wb-icon-link"
    LOCK_CLOSED = "wb-icon-lock-closed-2"
    LOCK_CLOSED_1 = "wb-icon-lock-closed-1"
    LOCK_CLOSED_3 = "wb-icon-lock-closed-3"
    LOCK_OPEN = "wb-icon-lock-open-2"
    LOCK_OPEN_1 = "wb-icon-lock-open-1"
    LOCK_OPEN_3 = "wb-icon-lock-open-3"
    LUNCH = "wb-icon-lunch"
    MALE = "wb-icon-male"
    MAN_1 = "wb-icon-man-1"
    MAN_2 = "wb-icon-man-2"
    MAN_3 = "wb-icon-man-3"
    MAN_4 = "wb-icon-man-4"
    MAN_5 = "wb-icon-man-5"
    MAN_6 = "wb-icon-man-6"
    MAN_7 = "wb-icon-man-7"
    MAXIMIZE = "wb-icon-maximize-3"
    MAXIMIZE_1 = "wb-icon-maximize-1"
    MAXIMIZE_2 = "wb-icon-maximize-2"
    MENU = "wb-icon-data-list-1"
    MINIMIZE = "wb-icon-minimize-4"
    MINIMIZE_1 = "wb-icon-minimize-1"
    MINIMIZE_2 = "wb-icon-minimize-2"
    MINIMIZE_3 = "wb-icon-minimize-3"
    NOTEBOOK = "wb-icon-notebook"
    NOTEBOOK_INVERTED = "wb-icon-notebook-inverted"
    NUMBERED_LIST = "wb-icon-numbered-list"
    PAGE_BREAK = "wb-icon-page-break"
    PENCIL = "wb-icon-edit-1"
    PEOPLE = "wb-icon-people"
    PERSON = "wb-icon-person"
    PHONE = "wb-icon-phone"
    PHONE_CALL = "wb-icon-phone-call"
    PIN = "wb-icon-pin"
    PIPETTE_EMPTY = "wb-icon-pipette-empty"
    PIPETTE_FULL = "wb-icon-pipette-full"
    PLUS = "wb-icon-plus"
    QUOTE = "wb-icon-quote"
    RECORDER = "wb-icon-recorder"
    REDO = "wb-icon-redo"
    REFRESH_1 = "wb-icon-refresh-1"
    REFRESH_2 = "wb-icon-refresh-2"
    REFRESH_24_ACW = "wb-icon-refresh-24-acw"
    REFRESH_24_CW = "wb-icon-refresh-24-cw"
    REFRESH_3_ACW = "wb-icon-refresh-3-acw"
    REFRESH_3_CW = "wb-icon-refresh-3-cw"
    REFRESH_CLOCK = "wb-icon-refresh-clock"
    REFRESH_DOTS = "wb-icon-refresh-dots"
    REFRESH_DOTS_24 = "wb-icon-refresh-dots-24"
    REFRESH_DOTS_GEAR = "wb-icon-refresh-dots-gear"
    REFRESH_THREE = "wb-icon-refresh-three"
    REMOVE_FORMATTING = "wb-icon-remove-formatting"
    REPLACE = "wb-icon-arrows-replace"
    SCISSORS = "wb-icon-scissors"
    SEARCH = "wb-icon-search"
    SEND = "wb-icon-send"
    SICK_LEAVE_1 = "wb-icon-sick-leave-1"
    SICK_LEAVE_2 = "wb-icon-sick-leave-2"
    SICK_LEAVE_3 = "wb-icon-sick-leave-3"
    SICK_LEAVE_4 = "wb-icon-sick-leave-4"
    SHARE_1 = "wb-icon-share-1"
    SHARE_2 = "wb-icon-share-2"
    SKYPE = "wb-icon-skype"
    SORT_A_Z = "wb-icon-from-a-to-z"
    SORT_Z_A = "wb-icon-from-z-to-a"
    SOUND_OFF = "wb-icon-sound-off"
    SOUND_ON = "wb-icon-sound-on"
    SPEACH_BUBBLE = "wb-icon-speech-bubble"
    SPELL_CHECK = "wb-icon-spell-check"
    STAR = "wb-icon-star"
    STATS = "wb-icon-chart-line-1"
    STEP_BACK_2 = "wb-icon-step-back-2"
    STEP_FORWARD_2 = "wb-icon-step-forward-2"
    STEP_LEFT = "wb-icon-step-back-1"
    STEP_RIGHT = "wb-icon-step-forward-1"
    STRIKETHROUGH = "wb-icon-strikethrough"
    SUBSCRIPT = "wb-icon-subscript"
    SUITCASE_1 = "wb-icon-suitcase-1"
    SUITCASE_2 = "wb-icon-suitcase-2"
    SUPERSCRIPT = "wb-icon-superscript"
    TABLE = "wb-icon-table"
    TEXT_ALIGN_CENTER = "wb-icon-text-align-center"
    TEXT_ALIGN_JUSTIFY = "wb-icon-text-align-justify"
    TEXT_ALIGN_LEFT = "wb-icon-text-align-left"
    TEXT_ALIGN_RIGHT = "wb-icon-text-align-right"
    TEXT_BG_COLOR = "wb-icon-text-bg-color"
    THUMBS_DOWN = "wb-icon-thumb-down"
    THUMBS_UP = "wb-icon-thumb-up"
    TRADE = "wb-icon-trade"
    TRASH = "wb-icon-trashbin-2"
    TRASHBIN_1 = "wb-icon-trashbin-1"
    TRIANGLES_DOWN = "wb-icon-triangles-down"
    TRIANGLES_LEFT = "wb-icon-triangles-left"
    TRIANGLES_LEFT_STOP = "wb-icon-triangles-left-stop"
    TRIANGLES_RIGHT = "wb-icon-triangles-right"
    TRIANGLES_RIGHT_STOP = "wb-icon-triangles-right-stop"
    TRIANGLES_UP = "wb-icon-triangles-up"
    TRIANGLE_DOWN = "wb-icon-triangle-down"
    TRIANGLE_LEFT = "wb-icon-triangle-left"
    TRIANGLE_RIGHT = "wb-icon-triangle-right"
    TRIANGLE_UP = "wb-icon-triangle-up"
    UNDERLINE = "wb-icon-underline"
    UNDO = "wb-icon-undo"
    UNLINK = "wb-icon-unlink"
    WOMAN_1 = "wb-icon-woman-1"
    WOMAN_2 = "wb-icon-woman-2"
    WOMAN_3 = "wb-icon-woman-3"
    WOMAN_4 = "wb-icon-woman-4"
    WOMAN_5 = "wb-icon-woman-5"
    WOMAN_6 = "wb-icon-woman-6"
    WOMAN_7 = "wb-icon-woman-7"
    X = "wb-icon-x"



from typing import Dict

import requests
from django.utils.html import strip_tags
from rest_framework.permissions import AllowAny
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.reverse import reverse

from bridger import display as dp
from bridger import serializers
from bridger.enums import Button, Operator, Unit, WBIcon
from bridger.settings import bridger_settings
from bridger.viewsets import ViewSet

story_type_choices = (("bug", "Bug"), ("feature", "Feature"))


def config(request: Request) -> Dict:
    return {
        "endpoint": reverse("bridger:clubhouse-list", request=request),
        "add": reverse("bridger:clubhouse-list", request=request),
    }


class ClubhouseSerializer(serializers.Serializer):
    id = serializers.PrimaryKeyField()
    name = serializers.CharField(label="Name")
    story_type = serializers.ChoiceField(label="Type", choices=story_type_choices)
    description = serializers.MarkdownTextField(required=False)
    created_at = serializers.DateTimeField(read_only=True)
    completed = serializers.BooleanField(read_only=True)

    def validate(self, data):
        data["name"] = strip_tags(data["name"])
        description = f"**Reported by: {self.context['request'].user.username}**\n\n{data['description']}"
        data["description"] = description

        return data

    def create(self, validated_data):
        assert bridger_settings.CLUBHOUSE_API_TOKEN
        assert bridger_settings.CLUBHOUSE_PROJECT_ID

        response = requests.post(
            f"https://api.clubhouse.io/api/v3/stories?token={bridger_settings.CLUBHOUSE_API_TOKEN}",
            json={**validated_data, "project_id": bridger_settings.CLUBHOUSE_PROJECT_ID,},
            headers={"Content-Type": "application/json"},
        )

        return self.__class__(response.json())


class ClubHouseView(ViewSet):
    IDENTIFIER = "clubhouse"
    LIST_ENDPOINT = "bridger:clubhouse-list"
    INSTANCE_ENDPOINT = "bridger:clubhouse-list"
    CREATE_ENDPOINT = "bridger:clubhouse-list"

    LIST_TITLE = "Bug Reports"
    INSTANCE_TITLE = "Bug Report"

    BUTTONS = [Button.REFRESH.value, Button.NEW.value]
    CREATE_BUTTONS = [Button.SAVE.value]

    INSTANCE_DISPLAY = dp.InstanceDisplay(sections=(dp.Section(fields=dp.FieldSet(("name", "story_type", "description"))),))

    LIST_DISPLAY = dp.ListDisplay(
        fields=[
            dp.Field(key="id", label="Identifier", width=Unit.FRACTION(1)),
            dp.Field(key="created_at", label="Created", width=Unit.FRACTION(1)),
            dp.Field(key="name", label="Name", width=Unit.FRACTION(8)),
        ],
        formatting=[
            dp.Formatting(
                column="story_type",
                formatting_rules=[
                    dp.FormattingRule(icon=WBIcon.BOLD.value, condition=dp.Condition(operator=Operator.EQUAL, value="bug"),),
                    dp.FormattingRule(
                        icon=WBIcon.DATA.value, condition=dp.Condition(operator=Operator.EQUAL, value="feature"),
                    ),
                ],
            ),
            dp.Formatting(
                column="completed",
                formatting_rules=[
                    dp.FormattingRule(style={"backgroundColor": "#228B22"}, condition=dp.Condition(Operator.EQUAL, value=True),)
                ],
            ),
        ],
        legends=[
            dp.Legend(
                items=[
                    dp.LegendItem(icon=WBIcon.BOLD.value, label="Bug"),
                    dp.LegendItem(icon=WBIcon.DATA.value, label="Feature"),
                ]
            )
        ],
    )

    serializer_class = ClubhouseSerializer

    # We need to add this otherwise DjangoModelPermissions complains it cannot find a get_queryset method
    permission_classes = (AllowAny,)

    def create(self, request):
        serializer = self.serializer_class(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)
        clubhouse_story = serializer.save()
        return Response({"instance": clubhouse_story.data}, status=201)

    def retrieve(self, request, pk=None):
        assert bridger_settings.CLUBHOUSE_API_TOKEN

        response = requests.get(
            f"https://api.clubhouse.io/api/v3/stories/{pk}?token={bridger_settings.CLUBHOUSE_API_TOKEN}",
            headers={"Content-Type": "application/json"},
        )

        if response.status_code == 404:
            return Response({}, status=404)
        json_response = response.json()
        serializer = self.serializer_class(json_response)
        return Response({"instance": serializer.data})

    def list(self, request):
        assert bridger_settings.CLUBHOUSE_API_TOKEN
        assert bridger_settings.CLUBHOUSE_PROJECT_ID

        response = requests.get(
            f"https://api.clubhouse.io/api/v3/projects/{bridger_settings.CLUBHOUSE_PROJECT_ID}/stories?token={bridger_settings.CLUBHOUSE_API_TOKEN}",
            headers={"Content-Type": "application/json"},
        )
        json_response = response.json()

        serializer = self.serializer_class(json_response, many=True)
        return Response({"results": serializer.data})



from dataclasses import dataclass
from typing import List, Tuple


@dataclass(unsafe_hash=True)
class FieldSet:
    fields: Tuple

    def __post_init__(self):
        assert isinstance(self.fields[0], (str, FieldSet)), "fields can only contain strings or more FieldSets."
        assert isinstance(self.fields, tuple), f"fields have to be a tuple, not {type(self.fields)}"

    def __iter__(self):
        for field in self.fields:
            if isinstance(field, str):
                yield field
            else:
                yield tuple(field)


@dataclass(unsafe_hash=True)
class SectionList:
    key: str

    def __iter__(self):
        yield "key", self.key


@dataclass(unsafe_hash=True)
class Section:
    fields: FieldSet = None
    section_list: SectionList = None
    title: str = None
    collapsed: bool = False

    def __post_init__(self):
        assert bool(self.fields) != bool(self.section_list), "Either fields have to be provided or section_list, not both."

    def __iter__(self):
        if self.fields:
            yield "fields", tuple(field for field in self.fields)
        elif self.section_list:
            yield "list", dict(self.section_list)

        yield "collapsed", self.collapsed

        if self.title:
            yield "title", self.title


@dataclass(unsafe_hash=True)
class InstanceDisplay:
    sections: Tuple[Section]

    def __post_init__(self):
        assert isinstance(self.sections, tuple), f"sections have to be a tuple, not {type(self.sections)}"

    def __iter__(self):
        # All sections need to be iterated over and be converted into a dict
        for section in self.sections:
            yield dict(section)



from bridger import serializers

from .models import Tag


class TagRepresentationSerializer(serializers.RepresentationSerializer):
    class Meta:
        model = Tag
        fields = ("id", "title", "color")


class TagSerializerMixin(serializers.ModelSerializer):
    tags = serializers.PrimaryKeyRelatedField(queryset=Tag.objects.all(), label="Tags", many=True)
    _tags = TagRepresentationSerializer(source="tags", many=True)



from slugify import slugify

from bridger.metadata import BridgerMetadata


class PandasMetadata(BridgerMetadata):
    def determine_metadata(self, request, view):
        metadata = super().determine_metadata(request, view)
        metadata["identifier"] += f"-{slugify(view.__class__.__name__)}"
        metadata["fields"] = view.pandas_fields.to_dict()
        return metadata



import logging
import os
from typing import Dict, List

from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.module_loading import import_string
from markdown.extensions.tables import TableExtension
from markdown_blockdiag import BlockdiagExtension
from rest_framework.request import Request
from rest_framework.reverse import reverse

from bridger.fsm.markdown_extensions import FSMExtension

logger = logging.getLogger(__name__)

DEFAULTS = {
    "PROFILE": "bridger.profile.default_profile",
    "DEFAULT_FRONTEND_USER_CONFIGURATION_ORDER": ["config__order"],
    "DEFAULT_AUTH_CONFIG": "bridger.auth.jwt_auth",
    "DEFAULT_NOTIFICATION_CONFIG": "bridger.notifications.settings.notification_config",
    "DEFAULT_SHARE_BUTTON": "bridger.share.buttons.share_action_button",
    "DEFAULT_SHARE_NOTIFICATION": "bridger.share.notifications.share_notification",
    "DEFAULT_SHARE_SERIALIZER": "bridger.share.serializers.ShareSerializer",
    "DEFAULT_MARKDOWN_EXTENSIONS": [TableExtension(), FSMExtension(), BlockdiagExtension(format="svg"),],
    "ADDITIONAL_DEFAUL_MODELVIEW_ATTRIBUTES": [],
    "FRONTEND_TEMPLATE": "bridger/frontend.html",
    "FRONTEND_CONTEXT": {
        "TITLE": "Workbench",
        "FONT_URL": "https://fonts.googleapis.com/css?family=Roboto:400,400i,500,500i,700,900&display=swap",
        "CDN_URLS": [
            "https://cdn.plot.ly/plotly-latest.min.js",
            "https://unpkg.com/react@16/umd/react.production.min.js",
            "https://unpkg.com/react-dom@16/umd/react-dom.production.min.js",
        ],
        "FAVICON_URL": None,
        "CSS_URL": None,
        "CONFIG_URL": None,
        "JS_URL": None,
    },
    "FRONTEND_MENU_CALENDAR": None,
    "MARKDOWN_TEMPLATE_TAGS": [],
    "CLUBHOUSE_CONFIG": "bridger.clubhouse.config",
    "CLUBHOUSE_API_TOKEN": os.environ.get("CLUBHOUSE_API_TOKEN", ""),
    "CLUBHOUSE_PROJECT_ID": os.environ.get("CLUBHOUSE_PROJECT_ID", ""),
}

IMPORT_STRINGS = [
    "PROFILE",
    "DEFAULT_AUTH_CONFIG",
    "DEFAULT_NOTIFICATION_CONFIG",
    "DEFAULT_SHARE_BUTTON",
    "DEFAULT_SHARE_NOTIFICATION",
    "DEFAULT_SHARE_SERIALIZER",
    "CLUBHOUSE_CONFIG",
]


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for API setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e,)
        raise ImportError(msg)


class BridgerSettings:
    """The settings, which is mostly taken from the settings module of Django Rest Framework"""

    def __init__(self, defaults=None, import_strings=None):
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS

    @property
    def settings(self):
        return getattr(settings, "BRIDGER_SETTINGS", {})

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError(f"Invalid Bridger Settings: {attr}")

        val = self.settings.get(attr, self.defaults[attr])

        if isinstance(val, dict):
            self.defaults[attr].update(val)
            val = self.defaults[attr]

        if attr in self.import_strings:
            val = perform_import(val, attr)

        return val


bridger_settings = BridgerSettings(DEFAULTS, IMPORT_STRINGS)



import pytest
from .mixins import TestModelClass, TestSerializerClass, TestrepresentationViewSetClass, TestViewSetClass
from .utils import get_all_subclasses
from bridger import serializers
from bridger import viewsets
from django.db import models
from django.urls import get_resolver
get_resolver().url_patterns

models = filter(lambda x: not x.__module__.startswith(('bridger', 'django', 'rest_framework')), get_all_subclasses(models.Model))
serializers = filter(lambda x: "bridger" not in x.__module__, get_all_subclasses(serializers.ModelSerializer))
representationviewsets = filter(lambda x: "bridger" not in x.__module__, get_all_subclasses(viewsets.RepresentationViewSet))
modelviewsets = filter(lambda x: "bridger" not in x.__module__, get_all_subclasses(viewsets.ModelViewSet))

@pytest.mark.django_db 
class TestProject:
    @pytest.mark.parametrize("model", models)
    def test_models(self, model):
        # print(mvs.__name__)
        # print(mvs.__module__)
        my_test = TestModelClass(model)
        my_test.execute_test()

    @pytest.mark.parametrize("serializer", serializers)
    def test_serializers(self, serializer):
        my_test = TestSerializerClass(serializer)
        my_test.execute_test()

    @pytest.mark.parametrize("rvs", representationviewsets)
    def test_representationviewsets(self, rvs, admin_client):
        my_test = TestrepresentationViewSetClass(rvs)
        my_test.execute_test(admin_client)

    @pytest.mark.parametrize("mvs", modelviewsets)
    def test_modelviewsets(self, mvs, admin_client):
        my_test = TestViewSetClass(mvs)
        my_test.execute_test(admin_client)


from dataclasses import dataclass, field
from typing import Callable, Dict, List, Optional, Union

from django.utils.http import urlencode
from rest_framework.request import Request
from rest_framework.reverse import reverse


@dataclass
class ItemPermission:
    permissions: List[str] = field(default_factory=list)
    method: Optional[Callable] = None

    def has_permission(self, request: Request) -> bool:
        if request.user.is_superuser:
            return True

        for permission in self.permissions:
            if not request.user.has_perm(permission):
                return False

        if self.method:
            return self.method(request=request)

        return True


@dataclass
class MenuItem:
    label: str

    endpoint: str
    endpoint_args: List[str] = field(default_factory=list)
    endpoint_kwargs: Dict[str, str] = field(default_factory=dict)
    endpoint_get_parameters: Dict[str, str] = field(default_factory=dict)
    reverse: bool = True

    permission: Optional[ItemPermission] = None
    add: Optional["MenuItem"] = None

    def __iter__(self):
        request = getattr(self, "request", None)
        if self.permission is None or self.permission.has_permission(request=request):
            if self.reverse:
                endpoint = reverse(
                    viewname=self.endpoint, args=self.endpoint_args, kwargs=self.endpoint_kwargs, request=request,
                )
            else:
                endpoint = self.endpoint

            if self.endpoint_get_parameters:
                endpoint += f"?{urlencode(self.endpoint_get_parameters)}"

            yield "label", self.label
            yield "endpoint", endpoint
            if self.add:
                self.add.request = request
                yield "add", dict(self.add)


@dataclass
class Menu:
    label: str
    items: List[Union[MenuItem, "Menu"]] = field(default_factory=list)

    index: Optional[int] = None

    def __iter__(self):
        request = getattr(self, "request", None)
        items = list()
        for item in filter(lambda x: bool(x), self.items):
            item.request = request
            serialized_item = dict(item)
            if serialized_item:
                items.append(serialized_item)

        if len(items) > 0:
            yield "label", self.label
            yield "items", items



from typing import Dict, Iterable


def uniquify_dict_iterable(iterable: Iterable[Dict], unique_key: str) -> Iterable[Dict]:
    keys = list()
    for item in iterable:
        if key := item.get(unique_key):
            if key not in keys:
                keys.append(key)
                yield item
        else:
            yield item



from typing import Dict

from bridger.metadata.mixins import BridgerViewSetConfig
from bridger.filters import DjangoFilterBackend
from bridger.utils import ilen

from rest_framework.filters import SearchFilter

class SearchFieldsBridgerViewSetConfig(BridgerViewSetConfig):

    def get_metadata(self) -> str:
        if SearchFilter in getattr(self.view, "filter_backends", []):
            return self.view.search_fields

    @classmethod
    def get_metadata_key(cls):
        return "search_fields"



from bridger.endpoints.metadata_config import EndpointConfig
from rest_framework.reverse import reverse

class NotificationEndpointConfig(EndpointConfig):
    def get_endpoint(self):
        return None

    def get_create_endpoint(self):
        return None
        
    def get_list_endpoint(self):
        return reverse("bridger:notification-list", request=self.request)

    def get_instance_endpoint(self):
        if self.instance:
            return None
        return self.get_list_endpoint()





import json
from pygments import highlight
from pygments.lexers import JsonLexer
from pygments.formatters import TerminalFormatter


class PrintColor:
    PURPLE = '\033[95m'
    CYAN = '\033[96m'
    DARKCYAN = '\033[36m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    END = '\033[0m'


def print_bold(text, color=None):
    if color:
        print(f"{PrintColor.BOLD}{color}{text}{PrintColor.END}")
    else:
        print(f"{PrintColor.BOLD}{text}{PrintColor.END}")


def print_dict(dictionary):
    print(highlight(json.dumps(dictionary, indent=4), JsonLexer(), TerminalFormatter()))



from typing import Optional

from bridger import display as dp
from bridger.display.metadata_config import DisplayConfig
from bridger.utils.icons import WBIcon

class NotificationDisplayConfig(DisplayConfig):

    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=(
                dp.Field(key="title", label="Title"),
                dp.Field(key="timestamp_created", label="Created"),
                dp.Field(key="message", label="Message"),
            ),
            formatting=[
                dp.Formatting(
                    column="timestamp_read", formatting_rules=[dp.FormattingRule(icon=WBIcon.EYE.value, condition=("∃", True)),],
                ),
            ],
        )

    def get_instance_display(self) -> Optional[dp.InstanceDisplay]:
        return dp.InstanceDisplay(
            sections=(
                dp.Section(
                    fields=dp.FieldSet(
                        fields=(
                            dp.FieldSet(fields=("timestamp_created", "timestamp_received")),
                            dp.FieldSet(fields=("timestamp_received", "timestamp_mailed")),
                            "title",
                            "message",
                        )
                    )
                ),
            )
        )



class BridgerFilterMixin:
    def __init__(self, *args, **kwargs):
        self.default = kwargs.pop("default", None)
        self.required = kwargs.pop("required", False)
        self.clearable = kwargs.pop("clearable", True)
        super().__init__(*args, **kwargs)

    def get_representation(self, request, name, view):
        self.key = name
        representation = {
            "label": self.label,
            "type": self.filter_type,
            "key": self.field_name,
            "lookup_expr": {self.lookup_expr: self.key},
            "clearable": self.clearable,
            "default": {},
        }

        # We consider the case where default is a boolean with value False.
        if not self.default == None:
            if callable(self.default):
                representation["default"][self.lookup_expr] = self.default(field=self, request=request, view=view)
            else:
                representation["default"][self.lookup_expr] = self.default  
        if self.required:
            representation["required"] = True
            assert representation["default"] != {}, "If a filter is required, it needs at least one default value"

        return representation



from typing import Any, Dict

from rest_framework.metadata import BaseMetadata
from rest_framework.request import Request
from rest_framework.views import View

from .mixins import BridgerViewSetConfig


class BridgerMetadata(BaseMetadata):

    def determine_metadata(self, request: Request, view: View) -> Dict[str, Any]:
        metadata = dict()

        # Check whether the current view is an instance view
        instance = "pk" in view.kwargs
        
        # Iterate over all subclasses that inherit from BridgerViewSetConfig
        for _class in BridgerViewSetConfig.__subclasses__(): 

            # Try to get the config class in one of two ways:
            # 1. Get it by accessing the instance field behind get_metadata_fieldname (Config classes that should/can be overriden)
            # 2. Use the subclasss (Config classes that should not be overriden)
            try:
                config_class = getattr(view, _class.get_metadata_fieldname())
                config = config_class(view=view, request=request, instance=instance)
            except NotImplementedError:
                config = _class(view=view, request=request, instance=instance)

            # Get the submetadata and if needed. 
            sub_metadata = getattr(config, _class.get_metadata_method())()
            try:
                metadata[_class.get_metadata_key()] = dict(sub_metadata)
            except (ValueError, TypeError) as e:
                metadata[_class.get_metadata_key()] = sub_metadata

        return metadata



from channels.generic.websocket import AsyncJsonWebsocketConsumer, JsonWebsocketConsumer


class AsyncAuthenticatedConsumerMixin:
    async def connect(self):
        user = self.scope.get("user", None)
        if user is not None:
            await self.accept()
        else:
            await self.close()


class AsyncAuthenticatedJsonWebsocketConsumer(AsyncAuthenticatedConsumerMixin, AsyncJsonWebsocketConsumer):
    pass



from bridger.titles.metadata_config import TitleConfig

class NotificationTitleConfig(TitleConfig):
    def get_instance_title(self):
        return "Notification: {{title}}"
    def get_list_title(self):
        return "Notifications"